    radix   dec
    #include P16F628A.INC

    ; ignore bogus warning messages when taking HIGH() of a
    ; negative number
    errorlevel -311 ; Operand of HIGH operator was > H'FFFF'

;===================================================================
;===================================================================
; PROGRAM NAME:     WWVB.ASM
; AUTHOR:           Bob Ammerman
;                   RAm Systems
; DATE:             22-May-2009
;
; This program may be freely used for any purpose, commercial or
; otherwise. There is no warranty of any kind associated with it.
; Use at your own risk.
;
; The program has been extensively tested under MPLAB/MPSIM, but
; has *not* been run in 'real hardware'.
;
; The author would appreciate an email describing your use of the
; program (or any bugs found) to: rammerman@verizon.net

;===================================================================
;===================================================================
; PURPOSE

; Simulates the WWVB time standard transmission. This program 
; generates the actual 60Khz carrier wave nearly identical to
; that generated by the WWVB time station (see LIMITATIONS, below).

;===================================================================
;===================================================================
; THEORY OF OPERATION

; The program works by generating 8 samples for each cycle of the
; 60 KHz carrier wave. This ends up being pretty close to a sine
; wave so little, if any, filtering should be required to remove
; harmonics from the signal.
;
;
; The following diagram shows how the samples are generated. Note
; that absolute value of the voltage at 45, 135, 225 and 315 
; degrees is 0.707 that at 90 and 270 degrees is 1.00

;
;               x
;
;         x           x
;                                            
;                                            
;
;   x-----------------------x-----------------------x
;  T00   T10   T20   T30   T40   T50   T60   T70   T80=T0 (instructions) 
;  000   045   090   135   180   225   270   315   360=0  (degrees)                         
;
;                                 x           x      
;
;                                       x            
;
;
; Each cycle of the 60Khz waveform takes exactly 80 instruction 
; times at the clock rate of 19.2MHz. This allows exactly 10 
; instructions to generate each sample. Two of these instructions 
; are used to actually output the sample, so we actually have only 
; 8 instruction times between samples. 

; There are many possible techniques that could be used to output
; the values of the sine wave samples. One simple choice would be
; an R-2R ladder type DAC connected to an output port. The 
; technique currently implemented in the program (but easily 
; changed) is outlined in the section "HARDWARE", below.

;===================================================================
;===================================================================
; HARDWARE

; As currently implemented the program is designed to run on a
; PIC16F628A. It can easily be ported to any PIC16 with a complete
; eight-bit port. 
;
; A group of 8 resistors are connected to the output pins to pull a 
; common output node up or down an appropriate amount at each 
; 'sample' of the output sine wave.

;          A = 39 Kohm
;    RB0  ---\/\/\/-----+     +V
;                       |      |
;          A = 39 Kohms |      |
;    RB1  ---\/\/\/-----+      |
;                       |      /
;          B = 56 Kohms |      \ Rup = 30 Kohm
;    RB2  ---\/\/\/-----+      /
;                       |      \
;          B = 56 Kohms |      |
;    RB3  ---\/\/\/-----+------+--------> output
;                       |      |
;         C = 360 Kohms |      /
;    RB4  ---\/\/\/-----+      \ Rdn = 30 Kohm
;                       |      /
;         C = 360 Kohms |      \
;    RB5  ---\/\/\/-----+      |
;                       |      |
;         D = 810 Kohms |    -----
;    RB6  ---\/\/\/-----+     ---
;                       |      -
;         D = 810 Kohms |
;    RB7  ---\/\/\/-----+
;
; See below for computation of the values of A, B, C and D
;
; The signal labelled 'output' above is relatively high 
; impedance and has a 1/2 +V bias, so it should be fed 
; through a voltage follower op-amp and then capacitively 
; coupled before being used. 

; It is easy to see that with all the RBx outputs turned 
; off the output voltage will be pulled by the two 30 Kohm
; resistors to 1/2 of +V. Assuming +V is 5V we thus have
; a 'quiescent point' of 2.5V

; We preset PORTB at initialization to have the value:
;
;   B'01010101'
; 
; Thus, when the corresponding bits of the TRIS register are
; cleared, the even numbered port bits are pullups and
; odd-numbered bits are pulldowns on the 'output' signal.
; On the other hand, when the TRIS bit is set then the pin 
; is an input and does not affect the output voltage at all.
;
; We generate the high-amplitude sine wave by enabling the 
; following pins with the TRIS register (note: the symbols 
; defined are used in the code that generates the waveform).

; (For the high-amplitude waveform we assume that with +V = 5V
;  that we want a 1V peak swing (2V pk-pk))

;                     ; ANGLE     PINS ENABLED   VOLTAGE CHANGE
;                     ; -----     ------------   -------------
HIAT0  = B'11111111'  ;   0        none            0.000 V
HIAT10 = B'11111110'  ;  45        RB0            +0.707 V
HIAT20 = B'11111010'  ;  90        RB0 & RB2      +1.000 V
HIAT30 = B'11111110'  ; 135        RB0            +0.707 V
HIAT40 = B'11111111'  ; 180        none            0.000 V
HIAT50 = B'11111101'  ; 225        RB1            -0.707 V
HIAT60 = B'11110101'  ; 270        RB1 & RB3      -1.000 V
HIAT70 = B'11111101'  ; 315        RB1            -0.707 V

; Note that we turn on a single resistor of A ohms to pull the
; output by 0.707V, and we turn on both an A ohm and a B ohm
; resistor to pull the output by 1.000V. This reduces the 
; drive requirement of the pin driving the B resistor vs. a 
; scheme which drove the A resistor for 0.707V and only the B 
; resistor for 1.000V. It also has the advantage of changing
; only one output pin at a time thus eliminating any possibility
; of a 'glitch'.

; Using a spreadsheet we get the following values for A and B:
;
; A = 38 Kohm, nearest standard value = 39 Kohm, error = 2.6%
; B = 55 Kohm, nearest standard value = 56 Kohm, error = 1.8%

; The low-amplitude waveform is 17dB down (power) from the 
; high-amplitude waveform. This is a factor of 50, so our 
; voltage swing for the low-amplitude waveform should be a 
; factor of sqrt(50) less than the swing for the high-amplitude
; waveform. So the peak swing is 1V/sqrt(50) = 0.141V. The swing
; at 45 degrees is 0.707V/sqrt(50) or 0.100V

; We generate the low-amplitude sign wave by enabling
; the following pins to drive the C and D resistors:

;                       ; ANGLE     PINS ENABLED   VOLTAGE CHANGE
;                       ; -----     ------------   --------------
LOAT0  = B'11111111'    ;   0       none            0.000 V
LOAT10 = B'11101111'    ;  45       RB4            +0.100 V       
LOAT20 = B'10101111'    ;  90       RB4 & RB6      +0.141 V
LOAT30 = B'11101111'    ; 135       RB4            +0.100 V
LOAT40 = B'11111111'    ; 180       none            0.000 V
LOAT50 = B'11011111'    ; 225       RB5            -0.100 V
LOAT60 = B'01011111'    ; 270       RB5 & RB7      -0.141 V
LOAT70 = B'11011111'    ; 315       RB5            -0.100 V

; Again, using a spreadsheet we get the following values for 
; C & D:
;
; C = 360 Kohm, nearest standard value = 360 Kohm, error = 0%
; D = 828 Kohm, nearest standard value = 810 Kohm, error = 2.2%

;===================================================================
;===================================================================
; LIMITATIONS

; Currently the program does not support:
;   Resetting the day-of-year at the end of the year
;   Incrementing the year field
;   The UT1 correction bits
;   The leap year bit
;   The leap second warning bit
;   The two DST flag bits
;
; The last four fifelds can be statically assigned any value
; and will then be sent with those values in every one minute 
; frame.
;
; No support exists for advancing the carrier by 45 degrees between
; 5 and 10 minutes after the hour. (This is used as a form of
; station identification by WWVB).

;===================================================================
;===================================================================
; PARAMETERS

; Define how fast time runs forward. The following symbol, which
; must be a multiple of 10 and greater than or equal to 20, defines 
; the number of 60Khz cycles per bit time (ie: second). Setting the 
; value to 60000 causes time to run at a normal 1 second per second.
; Setting it to a lower value will cause time to run faster by 
; generating fewer cycles of 60 KHz per bit time. For example, 
; setting it to 600 will result in time running 100 times faster 
; than real time. (Obviously the receive code trying to make sense 
; out of this would have to expect these much shorter bit times).

SPEED=60000 
;SPEED=60        ; for debugging, run 1000x faster
   
;===================================================================
;===================================================================
; Program variables

    udata   40h     ; make sure all the data is in BANK0!

; The current time. This is stored in BCD, with one digit per 
; byte.

T_MINLO     res 1
T_MINHI     res 1
T_HOURLO    res 1
T_HOURHI    res 1

; The Day-Of-Year bits are transmitted, and are incremented, but
; do not roll back to zero at the end of the year (they just keep
; incrementing past 365).

T_DAYLO     res 1
T_DAYMD     res 1
T_DAYHI     res 1

; The year bits are transmitted, but are not updated on the year
; rollover.

T_YEARLO    res 1
T_YEARHI    res 1

; The UT1 correction factor is transmitted from the following
; byte. Note that it is not updated by the code, but always
; remains constant.

T_UT1       res 1       ; UT1 correction factor
;   0 0 x x 0 0 0 0     ; 0 sec
;
;   0 1 x x 0 0 0 1     ; +.1 sec
;   0 1 x x 0 0 1 0     ; +.2 sec
;       ...     ; ...
;   0 1 x x 1 0 0 1     ; +.9 sec
;   
;   1 0 x x 0 0 0 1     ; -.1 sec
;   1 0 x x 0 0 1 0     ; -.2 sec
;       ...     ; ...
;   1 0 x x 1 0 0 1     ; -.9 sec

; Several additional flags are sent from the following byte.
; Note that they are not updated by the code, but always 
; remain constant. See the NIST documentation for more 
; details on what these bits do.

T_XXX       res 1   ; Miscellaneous bits
;   x x x x 1 x x x     ; Leap year flag
;   x x x x x 1 x x     ; Leap second warning bit
;   x x x x x x 1 x     ; DST flag bit 1
;   x x x x x x x 1     ; DST flag bit 2

; The loop counters for generating the high and low cycles

LO_LOOP_LSB res 1
LO_LOOP_MSB res 1
HI_LOOP_LSB res 1
HI_LOOP_MSB res 1

;===================================================================
;===================================================================
; RESET Vector

RESET code 0
    movlw   high(START)
    movwf   PCLATH
    goto    START

    code            ; let the linker place the rest of the code

;===================================================================
;===================================================================
; The OUTPUT macro MUST generate exactly two instructions to update
; the output value. It MUST not trash ANY state except W. In 
; particular it MUST NOT trash the flags in STATUS.
;
; Depending on the hardware configuration this macro could write a
; value to an output port, or to the TRIS register for the port.
; The currently assumed hardware (see the section HARDWARE, above
; assumes that we need to write the output values to the TRIS 
; register of port B. 
;
; Since the TRIS register is in bank 1 we point the FSR at it during
; initalization so we can refer to it as INDF while in bank 0 (which
; is where we store all our variables).

OUTPUT  MACRO  value
    MOVLW   value
    MOVWF   INDF
    ENDM

;===================================================================
;===================================================================
; The NOPX macro generates an instruction-cyle accurate delay of 
; up to 8 instruction times. 
; 
; The delay generated is exactly C2-C1+1 instruction cycles. This 
; convention allows the values passed as arguments to directy identify 
; the instruction cycles numbers (in the range 0 through 79) within 
; the 80 instruction cycles used to generate a single cycle of the
; output waveform.
;
; NOTE: An attempt to generate a delay longer than 8 instruction
;       times will result in an undefined symbol error for the
;       symbol NOP<n>, where n is the number of cycles.
;
; The generated code is as follows:
;
;     1 cycle     - NOP
;     2 cycles    - GOTO $+1
;     3 cycles    - NOP and GOTO $+1
;  >= 4 cycles    - CALL NOP<n>

NOPX    MACRO   C1,C2
    LOCAL   N,NEXT
N = C2-C1+1
  IF N < 4
    IF N&1
    NOP
    ENDIF
    IF N&2
    GOTO NEXT
NEXT:
    ENDIF
  ELSE
    CALL    NOP#V(N)
  ENDIF
    ENDM

; The following code is CALLed by the code generated by the NOPX 
; macro to delay the specified number of instructions (including 
; the call and return

NOP8:   NOP
NOP7:   NOP
NOP6:   NOP
NOP5:   NOP
NOP4:   RETURN

;===================================================================
;===================================================================
; The SENDBIT, SENDZERO and SENDMARK macros all expect to begin
; at instruction 76 within the 80 instructions allotted to generate
; each cycle of the output waveform. The next instruction following
; the expansion of each of these macros is also at instruction 76
; of a cycle. Thus, these macros should follow one another directly
; in the source code to generate a seamless output.

SENDBIT MACRO   REG,MASK    ;
    MOVF    REG,W           ;76 Get bit to send
    ANDLW   MASK            ;77 and mask it out
    CALL    ZERO_OR_ONE     ;78-79  send the bit
    ENDM

SENDZERO MACRO
    CLRW                    ;76 Force zero bit out
    ANDLW   0               ;77 Set Z flag
    CALL    ZERO_OR_ONE     ;78-79  send the bit
    ENDM

SENDMARK MACRO
    CALL    MARKER_78       ;76-77
    ENDM

;===================================================================
;===================================================================
; Initialize everything

START:
    ; make sure the bank select and indirect bank select bits are
    ; cleared
    clrf    STATUS

    ; disable the comparators so all the pins are digital
    banksel CMCON
    movlw   B'00000111'     
    movwf   CMCON

    ; initialize PORTB to the magic value B'01010101' as described in
    ; the section HARDWARE, above
    banksel PORTB
    movlw   B'01010101'
    movwf   PORTB
    
    ; the OUTPUT macro assumes that it can write to TRISB by writing
    ; to the INDF pseudo-register. We set FSR to point to TRISB here.
    movlw   TRISB           
    movwf   FSR             

    ; start all the pins set as inputs
    movlw   0FFh
    movwf   INDF

    ; make sure we have selected the bank with all the variables in
    ; it. the rest of the program assumes everything is reachable
    ; without further BANKSEL's
    banksel T_MINLO

    ; set the initial time
    movlw   0
    movwf   T_YEARHI
    movlw   9
    movwf   T_YEARLO

    movlw   1
    movwf   T_DAYHI
    movlw   9
    movwf   T_DAYMD
    movlw   9
    movwf   T_DAYLO

    movlw   2
    movwf   T_HOURHI
    movlw   3
    movwf   T_HOURLO

    movlw   5
    movwf   T_MINHI
    movlw   6
    movwf   T_MINLO

    ; set the extra values to reasonalble defaults
    clrf    T_UT1       ; no UT1 adjustment
    clrf    T_XXX       ; no leap second/leap year/daylight time

    GOTO    MAIN_LOOP

;===================================================================
;===================================================================
; Generate all the bits in a frame, except the final marker 
;
;   The 2 digit numbers preceded by "*" at the start of each comment 
;   are the second number within the minute (from 00 through 59)
;   
;   Each SENDMARK, SENDBIT and SENDZERO macro call uses exactly
;   one second worth of time, starting at instruction 76 before
;   the first cycle of low amplitude cycle for the bit, and 
;   ending with instruction 75 after the last high cycle for
;   the bit.

MAIN_LOOP:

    ; Output the first marker in the frame. We directly call the 
    ; MARKER_0 entry point to get the first output properly aligned.
    ; This is necessary because the GOTO MAIN_LOOP below consumes
    ; instructions 76-77 in the cycle and thus we can't user the 
    ; SENDMARK macro which assumes that it is running at cycle 76.

    CALL    MARKER_0      ;*00 78-79

    ; Output the middle 58 bits of the frame. Each output calls 
    ; one of the three macros SENDBIT, SENDZERO or SENDMARK, each
    ; of which takes exactly one second to execute.
    
    SENDBIT T_MINHI,4   ;*01 MS Digit of MINUTES
    SENDBIT T_MINHI,2   ;*02    
    SENDBIT T_MINHI,1   ;*03 

    SENDZERO            ;*04
    
    SENDBIT T_MINLO,8   ;*05 LS Digit of MINUTES
    SENDBIT T_MINLO,4   ;*06 
    SENDBIT T_MINLO,2   ;*07 
    SENDBIT T_MINLO,1   ;*08 

    SENDMARK            ;*09 

    SENDZERO            ;*10 
    SENDZERO            ;*11 

    SENDBIT T_HOURHI,2  ;*12 MS Digit of HOURS
    SENDBIT T_HOURHI,1  ;*13 

    SENDZERO            ;*14 

    SENDBIT T_HOURLO,8  ;*15 LS Digit of HOURS
    SENDBIT T_HOURLO,4  ;*16 
    SENDBIT T_HOURLO,2  ;*17 
    SENDBIT T_HOURLO,1  ;*18 

    SENDMARK            ;*19 
    
    SENDZERO            ;*20 
    SENDZERO            ;*21  

    SENDBIT T_DAYHI,2   ;*22 MS Digit of Day of year
    SENDBIT T_DAYHI,1   ;*23 

    SENDZERO            ;*24 

    SENDBIT T_DAYMD,8   ;*25 Middle Digit of Day of year
    SENDBIT T_DAYMD,4   ;*26 
    SENDBIT T_DAYMD,2   ;*27 
    SENDBIT T_DAYMD,1   ;*28 

    SENDMARK            ;*29 

    SENDBIT T_DAYLO,8   ;*30 LS Digit of Day of year
    SENDBIT T_DAYLO,4   ;*31 
    SENDBIT T_DAYLO,2   ;*32 
    SENDBIT T_DAYLO,1   ;*33 
    
    SENDZERO            ;*34 
    SENDZERO            ;*35

    SENDBIT T_UT1,40h   ;*36 UT1 Sign, +
    SENDBIT T_UT1,80h   ;*37 UT1 Sign, -
    SENDBIT T_UT1,40h   ;*38 UT1 Sign, +

    SENDMARK            ;*39
    SENDBIT T_UT1,8     ;*40 UT1 Correction value
    SENDBIT T_UT1,4     ;*41 
    SENDBIT T_UT1,2     ;*42 
    SENDBIT T_UT1,1     ;*43 

    SENDZERO            ;*44

    SENDBIT T_YEARHI,8  ;*45 MSDigit of Year
    SENDBIT T_YEARHI,4  ;*46 
    SENDBIT T_YEARHI,2  ;*47 
    SENDBIT T_YEARHI,1  ;*48 

    SENDMARK            ;*49

    SENDBIT T_YEARLO,8  ;*50 LSDigit of year
    SENDBIT T_YEARLO,4  ;*51
    SENDBIT T_YEARLO,2  ;*52
    SENDBIT T_YEARLO,1  ;*53 

    SENDZERO            ;*54 reserved bit

    SENDBIT T_XXX,8     ;*55 Leap year indicator
    SENDBIT T_XXX,4     ;*56 Leap second warning
    
    SENDBIT T_XXX,2     ;*57 Daylight savings time
    SENDBIT T_XXX,1     ;*58 Daylight savings time

; At this point we are at instruction 96 just before the first
; low-amplitude cycle of the final marker in the frame. We output
; the first low-amplitude cycle of the final marker inline here
; so that we can update the time variables for the next minute.

; This code is rather tricky. It has to bump the time while 
; 'simultaneously' keeping the output updated every 10 cycles.

    NOPX                     96,99

    ; output the first sample and increment the low digit of 
    ; the minutes and check for carry out

    OUTPUT  LOAT0           ;0,1
    INCF    T_MINLO,F       ;2  Bump LSDigit of minutes
    MOVF    T_MINLO,W       ;3  Load it for comparision
    XORLW   10              ;4  Set STATUS,Z if time to carry
    BTFSS   STATUS,Z        ;5  Are we carrying?
    GOTO    NOCARRY_8       ;6-7    No: we are done with this "bump"
    CLRF    T_MINLO         ;7  Zero after carry out
    NOPX                     8,9

    ; output the second sample and increment the high digit of
    ; the minutes and check for carry out

    OUTPUT  LOAT10          ;10-11
    INCF    T_MINHI,F       ;12 Bump MSDigit of minutes
    MOVF    T_MINHI,W       ;13 Load it for comparison
    XORLW   6               ;14 Set STATUS,Z if time to carry
    BTFSS   STATUS,Z        ;15 Are we carrying?
    GOTO    NOCARRY_18      ;16-17  No: we are done with this "bump"
    CLRF    T_MINHI         ;17 Zero after carry out
    NOPX                     18,19

    ; output the third sample and start processing the hours.
    ; if the pre-incremented value of the hours is less than 
    ; 20, then we know we can't carry out of hours into the 
    ; day-of-year. In this case we go to the code at HOURSLT20_25,
    ; below. Otherwise, we increment the ls digit of hours. If
    ; the result is less than 4 then we are done processing the 
    ; hours and don't need to carry to the day-of-year. Finally,
    ; if the result is 4 (ie: we have counted to hour 24), then
    ; we have to zero both halves of the hours and continue with
    ; incrementing the day of year.

    OUTPUT  LOAT20          ;20-21
    BTFSS   T_HOURHI,1      ;22 Are hours in [20..23]?
    GOTO    HOURSLT20_25    ;23-24  No: handle low hour count
    INCF    T_HOURLO,F      ;24 20->21, 21->22, 22->23, 23->24
    BTFSS   T_HOURLO,2      ;25 Did we carry from 23 to 24?
    GOTO    NOCARRY_28      ;26-27  No
    CLRF    T_HOURLO        ;27 Zero both halves of hours
    CLRF    T_HOURHI        ;28
    NOP                     ;29

    ; output the fourth sample and increment the low digit of the day-of-year
    OUTPUT  LOAT30          ;30-31  
    INCF    T_DAYLO,F       ;32 Bump LSDigit of day-of-year
    MOVF    T_DAYLO,W       ;33 Load it for comparison
    XORLW   10              ;34 Set STATUS,Z if time to carry
    BTFSS   STATUS,Z        ;35 Are we carrying?
    GOTO    NOCARRY_38      ;36-37  No: we are done with this "bump"
    CLRF    T_DAYLO         ;37 Zero after carry out
    NOPX                     38,39

    ; output the fifth sample and increment the middle digit of the day-of-year
    OUTPUT  LOAT40          ;40-41
    INCF    T_DAYMD,F       ;42 Bump middle digit of day-of-year
    MOVF    T_DAYMD,W       ;43 Load it for comparison
    XORLW   10              ;44 Set STATUS,Z if time to carry
    BTFSS   STATUS,Z        ;45 Are we carrying?
    GOTO    NOCARRY_48      ;46-47  No: we are done with this "bump"
    CLRF    T_DAYMD         ;47 Zero after carry out
    NOPX                     48,49                     

    ; output the sixth sample and increment the top digit of the day-of-year.
    OUTPUT  LOAT50          ;50-51
    INCF    T_DAYHI,F       ;52
    GOTO    NOCARRY_55      ;53-54

    ; the hours were less than 20, but we have to increment them.
    ; We know that we will not be carrying out of the hours field

HOURSLT20_25:
    NOPX                     25,29
    OUTPUT  LOAT30          ;30-31
    INCF    T_HOURLO,F      ;32 Bump LSdigit of hours
    MOVF    T_HOURLO,W      ;33 Load it for comparison
    XORLW   10              ;34 Set STATUS,Z if time to carry
    BTFSS   STATUS,Z        ;35 Are we carrying?
    GOTO    NOCARRY_38      ;36-37  No: we are done with this "bump"
    CLRF    T_HOURLO        ;37 zero after carry out
    INCF    T_HOURHI,F      ;38 ...and increment the high half of hours 
    NOP                     ;39
    OUTPUT  LOAT40          ;40-41
    GOTO    NOCARRY_44      ;42-43  We have no carry out of hours
        
NOCARRY_8:
    NOPX                     8,9
    OUTPUT  LOAT10          ;10-11
    NOPX                     12,17
NOCARRY_18:
    NOPX                     18,19
    OUTPUT  LOAT20          ;20-21
    NOPX                     22,27
NOCARRY_28:
    NOPX                     28,29
    OUTPUT  LOAT30          ;30-31
    NOPX                     32,37
NOCARRY_38:
    NOPX                     38,39
    OUTPUT  LOAT40          ;40-41
    NOPX                     42,43
NOCARRY_44:
    NOPX                     44,47
NOCARRY_48:
    NOPX                     48,49
    OUTPUT  LOAT50          ;50-51
    NOPX                     52,54
NOCARRY_55:
    NOPX                     55,59
    OUTPUT  LOAT60          ;60-61
    NOPX                     62,69
    OUTPUT  LOAT70          ;70-71
    NOPX                     72,77

    ; We are now ready to write the remaining 59999 cycles of the
    ; last marker bit. 
    CALL    MARKER_M1       ;78-79

    ; Continue with the next minute. Note that this GOTO is in 
    ; cycles 76-77, which means the first instruction in the
    ; MAIN_LOOP is at cycles 78-79, which is correct.
    GOTO    MAIN_LOOP       ;76-77

;===================================================================
;===================================================================
; Compute the high and low times for each bit type based on the 
; SPEED parameter

SPEED_D10 = SPEED / 10

N_Z_LO = SPEED_D10*2        ; The length of a zero bit low...
N_Z_HI = SPEED_D10*8        ; ...and high

N_O_LO = SPEED_D10*5        ; The length of a one bit low...
N_O_HI = SPEED_D10*5        ; ...and high

N_M_LO = SPEED_D10*8        ; The length of a mark bit low...
N_M_HI = SPEED_D10*2        ; ...and high

N_M1_LO = N_M_LO-1          ; The length of a mark bit low when
                            ; one cycle has been removed from it

;===================================================================
;===================================================================
; Output a MARKER bit
;   
;   The entry point MARKER_M1 generates one fewer low cycle.
;   the assumption is that the first cycle has already been
;   output by upper level code.
;
;   The entry point MARKER_0 assumes that the call branching
;   to it is in instruction times 78-79 within the cycle.
;
;   The entry point MARKER_78 assumes that the call branching
;   to it is in instruction times 76-77 within the cycle.
;   
;   The CALL to this routine should be in CYCLES 78-79
;   The first instruction after the call should be CYCLE 76

MARKER_M1:
    OUTPUT  LOAT0           ;0-1    output first sample of first cycle
    MOVLW   LOW(-N_M1_LO)   ;2  we are sending a marker less its
    MOVWF   LO_LOOP_LSB     ;3  minus its first cycle, thus we
    MOVLW   HIGH(-N_M1_LO)  ;4  output 47999 cycles of low 
    MOVWF   LO_LOOP_MSB     ;5  amplitude...
    GOTO    MARKER_8        ;6-7
MARKER_78:
    NOPX                     78,79  ; align properly
MARKER_0:
    OUTPUT  LOAT0           ;0-1    output first sample of first cycle
    MOVLW   LOW(-N_M_LO)    ;2  we are sending a marker, thus we
    MOVWF   LO_LOOP_LSB     ;3  output 48000 cycles of low 
    MOVLW   HIGH(-N_M_LO)   ;4  amplitude...
    MOVWF   LO_LOOP_MSB     ;5
    NOPX                     6,7
MARKER_8:
    NOPX                     8,9
    OUTPUT  LOAT10          ;10-11  (output 2nd sample of 1st cycle)
    MOVLW   LOW(-N_M_HI)    ;12 ...followed by 12000 cycles of
    MOVWF   HI_LOOP_LSB     ;13 high amplitude
    MOVLW   HIGH(-N_M_HI)   ;14
    MOVWF   HI_LOOP_MSB     ;15
    NOPX                     16,17
    GOTO    LO_LOOP_20      ;18-19

;===================================================================
; on entry: Z flag is set for a ZERO, clear for a ONE
;   
;   The CALL to this routine should be in CYCLES 78-79
;   The first instruction after the call should be CYCLE 76

ZERO_OR_ONE:        
    OUTPUT  LOAT0           ;0-1    output first sample of first cycle
    BTFSS   STATUS,Z        ;2  decide whether we are sending a 
    GOTO    ZERO_BIT        ;3-4    zero or a one
    NOP                     ;4

ONE_BIT:
    MOVLW   LOW(-N_O_LO)    ;5  we are sending a one, thus we
    MOVWF   LO_LOOP_LSB     ;6  output 30000 cycles of low
    MOVLW   HIGH(-N_O_LO)   ;7  amplitude...
    MOVWF   LO_LOOP_MSB     ;8
    NOP                     ;9
    OUTPUT  LOAT10          ;10-11  (output 2nd sample of 1st cycle)
    MOVLW   LOW(-N_O_HI)    ;12 ...followed by 30000 cycles of
    MOVWF   HI_LOOP_LSB     ;13 high amplitude
    MOVLW   HIGH(-N_O_HI)   ;14
    MOVWF   HI_LOOP_MSB     ;15
    NOPX                     16,17
    GOTO    LO_LOOP_20      ;18-19  continue with 3rd sample of 1st
                            ;   cycle

ZERO_BIT:
    MOVLW   LOW(-N_Z_LO)    ;5  we are sending a zero, thus we
    MOVWF   LO_LOOP_LSB     ;6  output 12000 cycles of low 
    MOVLW   HIGH(-N_Z_LO)   ;7  amplitude...
    MOVWF   LO_LOOP_MSB     ;8
    NOP                     ;9
    OUTPUT  LOAT10          ;10-11  (output 2nd sample of 1st cycle)
    MOVLW   LOW(-N_Z_HI)    ;12 ...followed by 48000 cycles of
    MOVWF   HI_LOOP_LSB     ;13 high amplitude
    MOVLW   HIGH(-N_Z_HI)   ;14
    MOVWF   HI_LOOP_MSB     ;15
    NOPX                     16,17
    GOTO    LO_LOOP_20      ;18-19  continue with 3rd sample of 1st
                            ;   cycle

LO_LOOP_0:
    OUTPUT  LOAT0           ;0-1    output first sample
    NOPX                     2,9
    OUTPUT  LOAT10          ;10-11  output second sample
    NOPX                     12,19

LO_LOOP_20:
    OUTPUT  LOAT20          ;20-21  output third sample
    INCF    LO_LOOP_MSB,F   ;22     assume carry from LSByte
    INCFSZ  LO_LOOP_LSB,F   ;23     increment LSByte
    DECF    LO_LOOP_MSB,F   ;24     carry didn't happen
    NOPX                     25,29
    OUTPUT  LOAT30          ;30-31  output fourth sample
    NOPX                     32,39
    OUTPUT  LOAT40          ;40-41  output fifth sample
    NOPX                     42,49
    OUTPUT  LOAT50          ;50-51  output sixth sample
    NOPX                     52,59
    OUTPUT  LOAT60          ;60-61  output seventh sample
    NOPX                     62,69
    OUTPUT  LOAT70          ;70-71  output eight sample
    NOP                     ;72
    CLRWDT                  ;73     don't let the dog bite
    MOVF    LO_LOOP_LSB,W   ;74     do we have more iterations?
    IORWF   LO_LOOP_MSB,W   ;75
    BTFSC   STATUS,Z        ;76
    GOTO    HI_LOOP_79      ;77-78  no: start high-amplitude cycles
    GOTO    LO_LOOP_0       ;78-79  yes: more low-amplitude cycles
  
HI_LOOP_79:
    NOP                     ;79

HI_LOOP_0:
    OUTPUT  HIAT0           ;0-1    output first sammple
    NOPX                     2,9
    OUTPUT  HIAT10          ;10-11  output second sample
    NOPX                     12,19
    OUTPUT  HIAT20          ;20-21  output third sample
    INCF    HI_LOOP_MSB,F   ;22     assume carry from LSByte
    INCFSZ  HI_LOOP_LSB,F   ;23     increment LSByte
    DECF    HI_LOOP_MSB,F   ;24     carry didn't happen
    NOPX                     25,29
    OUTPUT  HIAT30          ;30-31  output fourth sample
    NOPX                     32,39
    OUTPUT  HIAT40          ;40-41  output fifth sample
    NOPX                     42,49
    OUTPUT  HIAT50          ;50-51  output sixth sample
    NOPX                     52,59
    OUTPUT  HIAT60          ;60-61  output seventh sample
    MOVF    HI_LOOP_LSB,W   ;62     any more iterations?
    IORWF   HI_LOOP_MSB,W   ;63
    BTFSC   STATUS,Z        ;64
    GOTO    HI_EXIT         ;65-66  no: get ready to exit
    NOPX                     66,68
    CLRWDT                  ;69     don't let the dog bite
    OUTPUT  HIAT70          ;70-71  output eigth sample
    NOPX                     92,97
    GOTO    HI_LOOP_0       ;98-99  continue with next cycle

HI_EXIT:
    NOPX                     67,69
    OUTPUT  HIAT70          ;70-71
    NOPX                     72,73
    RETURN                  ;74,75
    
    END
    